"""Provides a common place for default values. Helps to avoid circular
importing."""
from typing import Literal

# Experiment Folders
EXP_DEF_DIR = "blueprints"
BATCH_DEF_DIR = "batch"
SHARED_MODEL_DIR = "models"

# Logging
LOG_DIR = "logs"
SCENARIO_LOGGER_NAME = "SCENARIO"