"""Common and useful outside of the module."""
from .structure import (
    BlueprintQuestion,
    FactoryOrder,    
    SupportedCaptureTypes,
    SupportedChoiceTypes,
    SupportsFactory
)
from .component import register_to_factory